# -*- coding: utf-8 -*-
"""
2022:10: 25, 30.

Reseni ulohy OPRBRN. 

Pocatecni reseni, ad hoc, A. 
"""



import pomocne as pm 
import numpy as np 
import math as mm



# # Nahrani ulohy OPRBRN resi soubor pomocne.py 
# # Zde pojmenujeme pracovni data:  
# xy = pm.xy 
# uzly = pm.uzly 
# hrany = pm.hrany 
# starty = pm.starty 
# blokace = pm.blokace 
# doby_oprav = pm.doby_oprav 
# legendaCG = pm.legendaCG 
# casyCG = pm.casyCG 
# blokaceCG = pm.blokaceCG 
# startyCG = pm.startyCG 
# uzlyKomp = pm.uzlyKomp 
# nazevUlohy = pm.nazevUlohy 
# uzlyKMG = pm.uzlyKMG 
# hranyKMG = pm.hranyKMG   # [odKomp1, doKomp2, dobaOpravy, idBlokaceCG] 
# Mgr = pm.Mgr



# Co je potreba udelat? 
# 1 - 4.4)  -- viz pomocne.py 
# 4.5) Generuj reseni ziskana Algoritmem Uspor. 
# 5..) Prejdi do souboru reseniM.py 


"""
Konstrukce pocatecniho reseni. 
Pouziva se savings algorithm (Clarke,Wright), ale ja pouziju jiny. 
Vyjdu z minimalni kostry grafu KMG. Kazda jednotka zacina v nejake stanici 
a postupne pridava hrany. Vybiraji si uzly tech blokaci, ktere jsou jim nejblize. 
Jednotky se stridaji. Pokud prave nepridavaji druhy konec blokace, pridavaji 
nejblizsiho souseda, tj. 1. uzel dalsi blokace. Pokracuji, dokud nevycerpaji 
vsechny hrany minimalni kostry. Tim je zajisteno pripustne reseni. 
"""



# 4.5: generovani pocatecnich reseni Algoritmem Uspor 
# Bude potreba zamyslet se nad modifikaci AU pro m-stanic. 
# Pokud provedu AU pouze nad krajnimi uzly blokaci z kostry site KMG, 
# tak tim ziskam nejen pripustne reseni, ale navic tim vyresim i (4.4). 

"""
    Vstupy: 
        casyCG... matice casu jizd mezi krajnimi uzly komponent nejkratsimi cestami  
        startyCG... indexy stanic v uplnem grafu CG 
        poctyTymuVeStanicich... z jedne stanice muze vystartovat vice tymu 
        navstivit... seznam seznamu dvojic uzlu z CG, ktere maji byt 
            navstiveny. Jedna se o dvojice, ktere maji byt navstiveny 
            bezprostredne po sobe oba uzly ze seznamu. Ma to vyznam 
            opraveni zablokovane hrany. 
    Prace: 
        je popsana v telech algoUspor 1, 2, 3(). 
    Vystupy: 
        reseni... dvojice (A,B), kde 
            A... je seznam seznamu indexu uzlu uplneho (CG) grafu, jinymi slovy 
                seznam tras. Seznam tras v poradi urcenem indexy startyCG a 
                poctyTymuVeStanicich. Kazdy zacina a konci ve stejne stanici. 
            B... je seznam seznamu nul a jednicek, pro kazdou hranu mezi 
                sousednimi uzly jedno cislo. Rika, ktera hrana ma byt opravena. 
    """



# AU pomocne funkce 

def au1rajony(casyCG, startyCG, poctyTymuVeStanicich): 
    ucg = np.arange(casyCG.shape[0]) 
    ucg = np.setdiff1d(ucg, startyCG) 
    # ucg obsahuje jen nestartovni uzly 
    rajony = [] 
    startyCGmod = startyCG[poctyTymuVeStanicich > 0] 
    poctyTymuVeStanicichMod = poctyTymuVeStanicich[poctyTymuVeStanicich > 0] 
    for i in range(len(startyCGmod)): 
        rajony.append([])
    # pridelime uzly z ucg jednotlivym rajonum (nejblizsim startum) 
    vzdalenostiZbytkuOdStanic = casyCG[np.ix_(startyCGmod,ucg)] 
    kam = np.argmin(vzdalenostiZbytkuOdStanic,0) 
    for i in range(len(startyCGmod)): 
        seznam = ucg[np.where(kam == i)[0]].tolist() 
        rajony[i] = seznam 
    vysledek = (startyCGmod, poctyTymuVeStanicichMod, rajony) 
    return(vysledek) 

# (startyCGmod, poctyTymuVeStanicichMod, rajony) = \
#     au3rajony(casyCG, startyCG, poctyTymuVeStanicich) 
def au3rajony(casyCG, startyCG, poctyTymuVeStanicich, navstivit): 
    ucg = set(np.array(navstivit).ravel().tolist()) 
    ucg = list(ucg - set(startyCG.tolist())) 
    # ucg obsahuje jen nestartovni uzly 
    rajony = [] 
    startyCGmod = startyCG[poctyTymuVeStanicich > 0] 
    poctyTymuVeStanicichMod = poctyTymuVeStanicich[poctyTymuVeStanicich > 0] 
    for i in range(len(startyCGmod)): 
        rajony.append([])
    # rajonum priradime dvojice z navstivit 
    for dvoj in navstivit: 
        rajvek = [] 
        for star in startyCGmod: 
            d = casyCG[star,dvoj[0]] + casyCG[star,dvoj[1]] 
            rajvek.append(d) 
        kde = np.argmin(np.array(rajvek)) 
        rajony[kde].append(dvoj) 
    vysledek = (startyCGmod, poctyTymuVeStanicichMod, rajony) 
    return vysledek 

def au_init(casyCG, startyCGmod, i, rajony): 
    startcg = startyCGmod[i] 
    rajon = rajony[i] 
    # inicializace 
    trasyr = []                  # trasy v rajonu 
    for uzelcg in rajon: 
        if type(uzelcg)==list :  # uzelcg je zde n-tice uzlu 
            trasar = [startcg] 
            for prvek in uzelcg: 
                trasar.append(prvek) 
            trasar.append(startcg) 
        else:                    # uzelcg je zde cislo jednoho uzlu 
            trasar = [startcg, uzelcg, startcg] 
        trasyr.append(trasar) 
    usporyr = np.zeros((len(trasyr),len(trasyr)),dtype=int) 
    znakyr = 11 * np.ones((len(trasyr),len(trasyr)),dtype=int) 
    for j in range(len(trasyr)-1): 
        for k in range(j+1, len(trasyr)): 
            (usp,znk) = uspora(casyCG, trasyr[j], trasyr[k]) 
            usporyr[j,k] = usp 
            usporyr[k,j] = usp   # matice uspor 
            znakyr[j,k] = znk 
            znakyr[k,j] = znk    # matice znaku 
    vysledek = (trasyr, usporyr, znakyr) 
    return vysledek

# trasa tr zde vzdy:  startA--uzly--startA
# znak: 11~[tr1,tr2], 12~[tr1,2rt], 21~[1rt,tr2], 22~[1rt,2rt] 
# znak znaci kombinaci orientacii tras, pri niz je dosazeno nejvetsi uspory. 
def uspora(casyCG, tr1:list, tr2:list): 
    znaky = [11,12,21,22] 
    ##tr1 = [381,191,192,381]   # 311: 0 1 2 3,  381: 191 192 209 220 
    ##tr2 = [381,209,220,381] 
    stan = tr1[0]        # stanice = tr1[0] = tr2[0] 
    tr1str = tr1[1:-1]   # stred 
    tr2str = tr2[1:-1] 
    a = casyCG[stan, tr1str[0]] 
    b = casyCG[stan, tr1str[-1]] 
    c = casyCG[stan, tr2str[0]] 
    d = casyCG[stan, tr2str[-1]] 
    h65 = casyCG[tr1str[-1], tr2str[0]] 
    h55 = casyCG[tr1str[0], tr2str[0]] 
    h66 = casyCG[tr1str[-1], tr2str[-1]] 
    h56 = casyCG[tr1str[0], tr2str[-1]] 
    rozdil11 = (b + c) - h65 
    rozdil12 = (b + d) - h66 
    rozdil21 = (a + c) - h55 
    rozdil22 = (a + d) - h56 
    rozdily = np.array([rozdil11,rozdil12,rozdil21,rozdil22]) 
    rozdil = np.max(rozdily)   # Nejvetsi rozdil je nase uspora. 
    znak = znaky[np.argmax(rozdily)] 
    return (rozdil, znak) 

# znak: 11~[tr1,tr2], 12~[tr1,2rt], 21~[1rt,tr2], 22~[1rt,2rt]
def sluc(tr1: list, tr2: list, znak: int):
    trasa = []
    stan = tr1[0]       # stanice = tr1[0] = tr2[0]
    tr1str = tr1[1:-1]  # stred
    tr2str = tr2[1:-1]
    trasa.append(stan)
    tr1strC = tr1str.copy()
    tr2strC = tr2str.copy()
    if (znak == 11):
        trasa.extend(tr1strC)
        trasa.extend(tr2strC)
    elif (znak == 12):
        trasa.extend(tr1strC)
        tr2strC.reverse()
        trasa.extend(tr2strC)
    elif (znak == 21):
        tr1strC.reverse()
        trasa.extend(tr1strC)
        trasa.extend(tr2strC)
    else:  # (znak==22)
        tr1strC.reverse()
        trasa.extend(tr1strC)
        tr2strC.reverse()
        trasa.extend(tr2strC)
    trasa.append(stan)
    return trasa

def krokShlukovani(casyCG, trasyr, usporyr, znakyr): 
    # Nutno zamezit vyberu diagonalnich prvku 
    india = np.arange(usporyr.shape[0]) 
    usporyr = np.array(usporyr, dtype=float) 
    usporyr[india,india] = - mm.inf 
    (idradku, idsloupce) = pm.argmatmax(usporyr)
    # V poradi urcenem pomoci odpovidajiciho znaku trasy shluknout 
    tr1 = trasyr[ idradku ] 
    tr2 = trasyr[idsloupce] 
    znak = znakyr[idradku, idsloupce] 
    trshl = sluc(tr1, tr2, znak)   # nova trasa 
    # Rekneme, ze se sloucila j-ta trasa s k-tou a vznikla nova. 
    # Z matice uspor i znaku smazat radky i sloupce j, k. 
    smazat = np.in1d(np.arange(usporyr.shape[0]), 
                     np.array([idradku,idsloupce])) 
    nechat = np.logical_not(smazat) 
    usporyr = usporyr[np.ix_(nechat,nechat)] 
    znakyr = znakyr[np.ix_(nechat,nechat)] 
    if tr1 in trasyr: 
        trasyr.remove(tr1) 
    if tr2 in trasyr: 
        trasyr.remove(tr2) 
    # Spocitat uspory i znaky pro novou trasu (trshl) a 
    usporyrn = [] 
    znakyrn = [] 
    for k in range(len(trasyr)): 
        (usp,znk) = uspora(casyCG, trshl, trasyr[k]) 
        usporyrn.append(usp) 
        znakyrn.append(znk) 
    # vznikle hodnoty pridat jako posledni radek i posledni sloupec 
    # do matice uspor respektive znaku. 
    usporyr = pm.obalMatici(usporyr, usporyrn, 0) 
    znakyr = pm.obalMatici(znakyr, znakyrn, 11) 
    # uz lze pridat novou trasu do tras rajonu 
    trasyr.append(trshl)  
    # Kolik je tras celkem? Pokud vic, nez poctyTymuVeStanicichMod[i], 
    # tak opakovat postup uvedeny vyse. 
    return (trasyr, usporyr, znakyr) 

""" Algoritmus uspor provest dvoufazove: 
1) Priradit uzly ke stanicim dle toho, ke kterym jsou nejblize. 
    A jen k neprazdnym stanicim, t.j. tem, ve kterych jsou nejaci cestari. 
2) Teprve pak delat shlukovani pro kazdou neprazdnou stanici zvlast. """ 



# Algoritmus uspor pro vsechny uzly uplneho grafu CG, tedy ne pro OPRBRN
def algoUspor1(casyCG:np.ndarray, startyCG:np.ndarray, 
              poctyTymuVeStanicich:np.ndarray): 
    trasy = [] 
    seznamSeznamuOprav = []   # budou same nuly 
    
    # 1) Prideleni uzlu k neprazdnym stanicim 
    (startyCGmod, poctyTymuVeStanicichMod, rajony) = \
        au1rajony(casyCG, startyCG, poctyTymuVeStanicich) 
    
    # 2) V i-tem rajonu poshlukujeme uzly do poctyTymuVeStanicichMod[i] tras 
    # uspora() aplikujeme jen na dvojice tras prislusnych jedne stanici 
    for i in range(len(startyCGmod)): 
        # inicializace 
        (trasyr, usporyr, znakyr) = au_init(casyCG, startyCGmod, i, rajony) 
        shlukovatDal = True 
        # Shlukovani: z matice uspor vybrat nejvetsi usporu 
        # a ji odpovidajici dvojici tras (pomoci argmatmax()) 
        while shlukovatDal: 
            (trasyr, usporyr, znakyr) = \
                krokShlukovani(casyCG, trasyr, usporyr, znakyr)
            if len(trasyr) <= poctyTymuVeStanicichMod[i]: 
                shlukovatDal = False 
        trasy.extend(trasyr)   # lisi se prvnim uzlem 
        for m in trasyr: 
            zadneOpravy = [0]*(len(m)-1) 
            seznamSeznamuOprav.append(zadneOpravy) 
    
    reseni = (trasy, seznamSeznamuOprav) 
    return reseni 

# # Test algoUspor1() 
# #poctyTymuVeStanicich = np.array([1,1,1,1,1]) 
# poctyTymuVeStanicich = np.array([2,0,3,0,0]) 
# #poctyTymuVeStanicich = np.array([1,2,3,4,5]) 
# pocatecniReseni = algoUspor1(casyCG,startyCG,poctyTymuVeStanicich) 



# Algoritmus uspor pro jednotlive uzly, avsak ne pro OPRBRN, 
# ale jen pro rozvoz materialu do komponent. (Netreba navstivit 
# vsechny uzly, ale vsechny komponenty. A nic se neopravuje.) 
# Dulezite je navstivit kazdou komponentu aspon jednou. 
def algoUspor2(casyCG:np.ndarray, startyCG:np.ndarray, 
              poctyTymuVeStanicich:np.ndarray, legendaCG:np.ndarray): 
    komponenty = [i for i in range(max(legendaCG[:,2])+1)] 
    uzlyVyvolene = set()   # ... ke konecnemu shlukovani 
    """ Z promenne komponenty budu odebirat navstivene komponenty. Jak? 
    - slucovani do tras musi byt provadeno stridave pres rajony, 
    chvili rajon1, chvili rajon2, rajon3,.. a opakovat. 
    Jinak se muze stat, ze propojeni komponent zaridi jen jedna stanice. 
    My ale chceme zamestnat vsechny tymy vsech stanic. """
    
    # 1) Prideleni uzlu k neprazdnym stanicim (vznik rajonu) 
    (startyCGmod, poctyTymuVeStanicichMod, rajony) = \
        au1rajony(casyCG, startyCG, poctyTymuVeStanicich) 
    # Inicializace 
    TRASYR = [] 
    USPORYR = [] 
    ZNAKYR = [] 
    for i in range(len(startyCGmod)): 
        (trasyr, usporyr, znakyr) = au_init(casyCG, startyCGmod, i, rajony) 
        TRASYR.append(trasyr) 
        USPORYR.append(usporyr) 
        ZNAKYR.append(znakyr) 
    
    # 2) Predbezne shlukovani 
    """ Budeme shlukovat uzly stridave ve vsech rajonech az do doby, 
    # kdy mezi komponentami nezustane zadna nenavstivena. 
    # uspora() aplikujeme jen na dvojice tras prislusnych jedne stanici """ 
    i = 0 
    while (len(komponenty) > 0):   # dokud nenavstivili vsechny k.: 
        trasyr = TRASYR[i % len(rajony)] 
        usporyr = USPORYR[i % len(rajony)] 
        znakyr = ZNAKYR[i % len(rajony)] 
        # # Pokud to jednou bude fungovat blbe, tak asi chybi toto: 
        # pt = poctyTymuVeStanicichMod[i % len(rajony)]
        pt = 1 
        if (len(trasyr) <= pt):  # neni-li co shlukovat: 
            i += 1               # jdeme do jineho rajonu 
            continue 
        """ Ty dve trasy z trasyr, kt shluknu, maji uprostred uzly, 
        ktere jsou kazdy v nejake komponente. Tyto uzly ulozim. 
        Tyto komponenty odmazu. """ 
        (trasyr, usporyr, znakyr) = \
            krokShlukovani(casyCG, trasyr, usporyr, znakyr) 
        TRASYR[i % len(rajony)] = trasyr 
        USPORYR[i % len(rajony)] = usporyr 
        ZNAKYR[i % len(rajony)] = znakyr 
        tr = trasyr[-1]   # nejnovejsi trasa 
        trstr = tr[1:-1]  # a jeji vnitrek 
        # trstr je list of ints, indexu do casyCG, kt jsou i v legendaCG[:,1] 
        idskomp = legendaCG[np.in1d(legendaCG[:,1],np.array(trstr)),2] 
        idskomp = np.unique(idskomp).tolist() 
        for komponenta in idskomp: 
            if komponenta in komponenty: 
                komponenty.remove(komponenta) 
        # pridame uzly do mnoziny uzlu vyvolenych ke konecnemu shlukovani 
        uzlyVyvolene = uzlyVyvolene | set(tr) 
        i += 1   # jdeme do jineho rajonu 
    uzlyVyvolene = list(uzlyVyvolene) 
    
    # 3) Konecne shlukovani    (nad zvolenymi uzly) 
    uzlyVyvolene = np.array(uzlyVyvolene) 
    nechat = np.in1d(np.arange(casyCG.shape[0]), uzlyVyvolene) 
    casyCGpodmatice = casyCG[np.ix_(nechat,nechat)] 
    startyCGmodp = [] 
    for q in startyCGmod: 
        p = np.where(uzlyVyvolene == q)[0][0] 
        startyCGmodp.append(p) 
    startyCGmodp = np.array(startyCGmodp) 
    poctyTymuVeStanicichModp = poctyTymuVeStanicichMod  # zustava stejne 
    reseni = algoUspor1(casyCGpodmatice, startyCGmodp, 
                  poctyTymuVeStanicichModp) 
    
    # reseni v puvodnim indexovani: 
    trasyNove = [] 
    opravy = reseni[1].copy() 
    for j in range(len(reseni[0])): 
        trasa = reseni[0][j] 
        trasaNova = [] 
        for k in trasa: 
            trasaNova.append(uzlyVyvolene[k]) 
        trasyNove.append(trasaNova) 
    
    reseniNove = (trasyNove, opravy) 
    return reseniNove 



""" Algoritmus uspor pro dvojice uzlu. 
Je-li do okruhu zvolen u1 z blokace, musi v nem byt i u2 z teze blokace. 
Lze pouzit k vypoctu pocatecniho reseni nad zvolenymi blokovanymi hranami. 
S dobami oprav tato funkce vubec nepocita, protoze doby oprav v algoritmu 
uspor nehraji zadnou roli. algoUspor4 vraci reseni s vyznacenymi opravenymi useky, 
coz jsou vsechny hrany dane seznamem "navstivit". """
def algoUspor3(casyCG:np.ndarray, startyCG:np.ndarray, 
              poctyTymuVeStanicich:np.ndarray, navstivit:list): 
    trasy = [] 
    seznamSeznamuOprav = [] 
    npnavstiv = np.array(navstivit) 
    
    # 1) Prideleni dvojic uzlu k neprazdnym stanicim 
    (startyCGmod, poctyTymuVeStanicichMod, rajony) = \
        au3rajony(casyCG, startyCG, poctyTymuVeStanicich, navstivit) 
    # Tentokrat rajony obsahuji kazdy nekolik dvojic uzlu z CG 
    
    # 2) V i-tem rajonu poshlukujeme uzly do poctyTymuVeStanicichMod[i] tras 
    # uspora() aplikujeme jen na dvojice tras prislusnych jedne stanici 
    for i in range(len(startyCGmod)): 
        if (len(rajony[i]) <= 0):           # je-li rajon prazdny: 
            for j in range(poctyTymuVeStanicichMod[i]): 
                trasy.append([startyCGmod[i]])  # nikam se nevyjelo 
                seznamSeznamuOprav.append([])   # zadnou hranou se neprojelo 
            continue                        # jdeme do jineho rajonu 
        
        # inicializace 
        (trasyr, usporyr, znakyr) = au_init(casyCG, startyCGmod, i, rajony) 
        shlukovatDal = True 
        # Shlukovani: z matice uspor vybrat nejvetsi usporu 
        # a ji odpovidajici dvojici tras (pomoci argmatmax()) 
        while shlukovatDal: 
            (trasyr, usporyr, znakyr) = \
                krokShlukovani(casyCG, trasyr, usporyr, znakyr) 
            if (len(trasyr)==1) or (len(trasyr) <= poctyTymuVeStanicichMod[i]): 
                shlukovatDal = False 
        trasy.extend(trasyr)   # lisi se prvnim uzlem 
        
        # identifikace blokaci a vyznaceni jejich oprav 
        for trasa in trasyr: 
            opravy = [] 
            for uid in range(len(trasa)-1): 
                shody1 = np.logical_and(npnavstiv[:,0]==trasa[uid], 
                               npnavstiv[:,1]==trasa[uid+1]) 
                shody2 = np.logical_and(npnavstiv[:,1]==trasa[uid], 
                               npnavstiv[:,0]==trasa[uid+1]) 
                shodyc = np.logical_or(shody1, shody2) 
                if any(shodyc): 
                    opravy.append(1)   #   oprava   blokace 
                else: 
                    opravy.append(0)   # neopravy neblokace 
            seznamSeznamuOprav.append(opravy) 
        
    reseni = (trasy, seznamSeznamuOprav) 
    return reseni 

# kostry = pm.levneKostry(multigraf=Mgr, pocetKoster=10) 
# kostraKMG = kostry[0] 
# # poctyTymuVeStanicich = np.array([1]*len(startyCG)) 
# # poctyTymuVeStanicich = np.array([2,0,3,0,0]) 
# # poctyTymuVeStanicich = np.array([1,2,3,4,5]) 
# navstivit = blokaceCG[kostraKMG[:,3],0:2].tolist() 
# # casyOpravNavs = blokaceCG[kostraKMG[:,3],2].tolist() 
# pocatecniReseni = algoUspor3(casyCG, startyCG, poctyTymuVeStanicich, navstivit) 


