# -*- coding: utf-8 -*-
"""
2022:12: 17, 
2023:01: 11, 
2023:03: 14 

Reseni ulohy OPRBRN. 

Hlavni reseni, mravenci, M. 
Zpusob konstrukce, ktery uziva a meni dostupnost uzluu. 
"""


import pomocne as pm 
import reseniA 
import reseniZ 
import numpy as np 
import math as mm 



# Nahrani ulohy OPRBRN resi soubor pomocne.py 
# Zde pojmenujeme pracovni data:  
xy = pm.xy 
uzly = pm.uzly 
hrany = pm.hrany 
starty = pm.starty 
blokace = pm.blokace 
doby_oprav = pm.doby_oprav 
legendaCG = pm.legendaCG 
casyCG = pm.casyCG 
blokaceCG = pm.blokaceCG 
startyCG = pm.startyCG 
uzlyKomp = pm.uzlyKomp 
uzlyKMG = pm.uzlyKMG 
hranyKMG = pm.hranyKMG   # [odKomp1, doKomp2, dobaOpravy, idBlokaceCG] 
Mgr = pm.Mgr_OJ   # Hrany ohodnoceny souctem dob Oprav i dob Jizd 

# kod blokovanych hran v neorientovanem grafu CG, klic do blokaceCG i hranyKMG 
blokaceCGkody = pm.kodyzHranCG(blokaceCG[:,0:2]) 
# konstrukceTras() pocita s cestami v CG 
cestyCG = pm.CESTY_CG 



""" V pripade jednoho mravence > 
Format reseni A je stary. Pise se: 
reseni = (trasy, seznamSeznamuOprav), napr.: 
reseni = ([10,18,12,10], [0,1,0]) 
ale v kontextu reseniM to znamena: 
reseni = (mravenciVolby, seznamSeznamuOprav) 
Od tohoto souboru pouzivam reseni ve formatu M a to znamena: 
reseni = (trasy, seznamSeznamuOprav, mravenciVolby), napr.: 
reseni = ([10,7,3,11,18,5,51,12,9,10], [0,0,0,0, 1,0,1, 0,0], [10,18,12,10]) nebo 
reseni = ([10,7,3,11,18,5,51,12,9,10], [0,0,0,0, 1,0,0, 0,0], [10,18,12,10]) 
Nastava ta potiz, pro nektera reseni ve formatu A nelze ziskat 
reseni ve formatu M jednoznacne, protoze M je podrobnejsi - jednu hranu CG 
lze rozvinout v cestu CG z nekolika hran, kde muze lezet vice blokaci, 
z cehoz nejde nijak usoudit, ktere byly konkretni cetou opraveny. 
adapterReseniAZ_M potrebuju kvuli pouziti funkce reseniA.algoUspor3(). 
reseniA.algoUspor3() dostane na vstup seznam nekterych blokovanych 
hran vybranych k oprave -tyhle tedy v seznamuSeznamuOprav budou mit 1. 
Ostatni CG hrany, ktere doplnim, budou mit 0. 
Lze pouzit i jako adapter reseni Z po k-optu zpet do formatu reseni M. 
adapterReseniZM(reseni) = adapterReseniAZ_M(reseni) 
"""
def adapterReseniAZ_M(reseni): 
    seznamTras = [] 
    seznamSeznamuOprav = [] 
    mravenciVolby = reseni[0] 
    oprneopr = reseni[1] 
    for m in range(len(mravenciVolby)):   # pro kazdeho mravence 
        trasa = [] 
        opravy = [] 
        start = mravenciVolby[m][0] 
        trasa.append(start) 
        for i in range(len(oprneopr[m])): 
            co = oprneopr[m][i]   # 0=neopravit, 1=opravit 
            if (co==1): 
                cil = mravenciVolby[m][i+1] 
                trasa.append(cil) 
                opravy.append(1) 
            else: # (co==0): 
                start = mravenciVolby[m][i] 
                cil = mravenciVolby[m][i+1] 
                cesta = cestyCG[start][cil] 
                trasa.extend(cesta[1:]) 
                opravy.extend([0]*(len(cesta)-1))
        seznamTras.append(trasa) 
        seznamSeznamuOprav.append(opravy) 
    return (seznamTras, seznamSeznamuOprav, mravenciVolby) 

def cenyTras(reseni:tuple): 
    return pm.cenyTras(pm.casyCG, reseni, pm.blokaceCG) 
# cenyTras(([[383]],[[]]))    # [0] 



def pripravParametry(vstupniParametry:dict, casyCG:np.ndarray, 
                     poctyTymuVeStanicich:np.ndarray):
    """ 
    Vstupy: 
        casyCG... Matice cen dopravy mezi uzly CG site. 
    Prace: 
        Pripravi zakladni hodnoty parametru ACO algoritmu, ktere lze zmenit. 
    Vystupy: 
        mravencu... Pocet mravencu v kolonii.
        iterLimit... maximalni pocet sezon. Sezona je obdobi, kdy kazdy z mravencu 
        zkonstruuje reseni a vrati se na svuj start. 
        odpar... cislo mezi 0 a 1. 0 - zadny feromon se neodpari. 1 - odpari se vse. 
        feroMin... dolni limit pro zustatek feromonu na hrane. Mene nebude.
        feroMax... horni limit pro zustatek feromonu na hrane. Vice nebude.
        feromony... matice obsahujici ohodnoceni hran feromony. 
        heuristika... matice obsahujici ohodnoceni hran heuristikou. 
        frekvencePruzkumu... Klesajici funkce jedne promenne. Nabyva hodnot 1 az 0. 
        Dava do vztahu pravdepodobnost uziti iteration-best solution proti 
        best-so-far solution v zavislosti na cisle iterace. 
        iterLimitKopt... maximalni pocet iteraci zlepsovaci heuristiky k-opt
        zlepsovaciAlgoritmus... bud 'opt2' nebo 'opt3', pokud 'nic', 
            tak se zlepsovaci algoritmus nepouzije. 
        typCeny... 'sum' pro soucet dob praci vsech tymu, 'max' pro dobu prace 
            tymu, kteremu prace trvala nejdele, 'vse' pro vektor dob praci vsech 
    """ 
    Ncg = casyCG.shape[0] 
    parametry = {} 
    
    parametry["mravencu"] = int(sum(poctyTymuVeStanicich)) 
    
    if ("odpar" in vstupniParametry.keys()): 
        parametry["odpar"] = vstupniParametry["odpar"] 
    else: 
        parametry["odpar"] = 0.2 
    
    if ("iterLimit" in vstupniParametry.keys()): 
        parametry["iterLimit"] = vstupniParametry["iterLimit"] 
    else: 
        parametry["iterLimit"] = 1000 
    
    if ("pomerFmaxFmin" in vstupniParametry.keys()): 
        parametry["pomerFmaxFmin"] = vstupniParametry["pomerFmaxFmin"] 
    else: 
        parametry["pomerFmaxFmin"] = 2 * Ncg 
    pomerFmaxFmin = parametry["pomerFmaxFmin"] 
    
    if ("feroMax" in vstupniParametry.keys()): 
        parametry["feroMax"] = vstupniParametry["feroMax"] 
    else: 
        parametry["feroMax"] = 1 
    feroMax = parametry["feroMax"]
    
    if ("feroMin" in vstupniParametry.keys()): 
        parametry["feroMin"] = vstupniParametry["feroMin"] 
    else: 
        parametry["feroMin"] = feroMax / (pomerFmaxFmin) 
    
    # Matice feromonu nesmi byt 0, jinak by ACO uvazlo hned na zacatku. 
    if ("feromony" in vstupniParametry.keys()): 
        parametry["feromony"] = vstupniParametry["feromony"] 
    else: 
        parametry["feromony"] = feroMax * np.ones([Ncg,Ncg]) 
    
    if ("alfa" in vstupniParametry.keys()): 
        parametry["alfa"] = vstupniParametry["alfa"] 
    else: 
        parametry["alfa"] = 1 
    
    if ("beta" in vstupniParametry.keys()): 
        parametry["beta"] = vstupniParametry["beta"] 
    else: 
        parametry["beta"] = 1 
    
    if ("frekvencePruzkumu" in vstupniParametry.keys()): 
        parametry["frekvencePruzkumu"] = vstupniParametry["frekvencePruzkumu"]
    else: 
        parametry["frekvencePruzkumu"] = lambda x: (0.5) 
        # lambda x: (1-mm.tanh((4*x-2*iterLimit)/iterLimit))/2 
    
    if ("zlepsovaciAlgoritmus" in vstupniParametry.keys()): 
        parametry["zlepsovaciAlgoritmus"] = vstupniParametry["zlepsovaciAlgoritmus"] 
    else: 
        parametry["zlepsovaciAlgoritmus"] = 'opt2'  
        # hodnoty {'opt2', 'opt3', 'nic'} 
        
    if ("iterLimitKopt" in vstupniParametry.keys()): 
        parametry["iterLimitKopt"] = vstupniParametry["iterLimitKopt"] 
    else: 
        parametry["iterLimitKopt"] = mm.inf 
    
    if ("uzitPocatecniReseni" in vstupniParametry.keys()): 
        parametry["uzitPocatecniReseni"] = vstupniParametry["uzitPocatecniReseni"] 
    else: 
        parametry["uzitPocatecniReseni"] = True 
    
    if ("typCeny" in vstupniParametry.keys()): 
        parametry["typCeny"] = vstupniParametry["typCeny"] 
    else: 
        parametry["typCeny"] = 'vse' 
        # hodnoty {'vse', 'sum', 'max'} 
    
    if ("zpusobVyberuMravence" in vstupniParametry.keys()): 
        parametry["zpusobVyberuMravence"] = vstupniParametry["zpusobVyberuMravence"] 
    else: 
        parametry["zpusobVyberuMravence"] = 'vm2'  
        # hodnoty {'vm1', 'vm2'} 
    
    if ("pocetKolonii" in vstupniParametry.keys()): 
        parametry["pocetKolonii"] = vstupniParametry["pocetKolonii"] 
    else: 
        parametry["pocetKolonii"] = 1 
    
    if ("aktualizovatFeromony" in vstupniParametry.keys()): 
        parametry["aktualizovatFeromony"] = vstupniParametry["aktualizovatFeromony"] 
    else: 
        parametry["aktualizovatFeromony"] = True 
    
    return parametry 



def vyberemeMravence1(trasy, odpracovaneDoby, feromony, alfa, beta, 
                      indexStaniceMravence, seznamyDostupnychUzlu, 
                      seznamDostupnychBlokaci): 
    """ Vyber mravence k presunu podle R.Vodak et al., 2018 """ 
    m = len(trasy) 
    char = np.array([0]*m, dtype=float) 
    kandidati = [None]*m 
    for a in range(m): 
        index_stanice = indexStaniceMravence[a] 
        aktualni = trasy[a][-1] 
        sousede = seznamyDostupnychUzlu[index_stanice] 
        if (len(sousede)==0): 
            continue 
        f = feromony[aktualni,sousede] 
        f = np.power(f, alfa) 
        h = casyCG[aktualni,sousede] 
        h = 1 / h 
        h = np.power(h, beta) 
        psti = f*h / f.dot(h) 
        cpsti = np.cumsum(psti) 
        trfal = (cpsti > np.random.rand())
        vyber = np.where(trfal)[0][0] 
        novy = sousede[vyber] 
        kandidati[a] = novy 
        cesta = cestyCG[aktualni][novy] 
        hranyCesty = np.vstack((np.array(cesta[0:-1]), np.array(cesta[1:]))) 
        hranyCesty = hranyCesty.transpose() 
        kodyCGhranCesty = pm.kodyzHranCG(hranyCesty) 
        hranaJeDostupnaBlokace = np.in1d(
            kodyCGhranCesty, seznamDostupnychBlokaci) 
        if any(hranaJeDostupnaBlokace): 
            opravovaneBlokace = kodyCGhranCesty[hranaJeDostupnaBlokace] 
            opravovanaBlokace = opravovaneBlokace[0]  # jen prvni 
            ktbl = np.in1d(blokaceCGkody, opravovanaBlokace) 
            dobaOprav = blokaceCG[ktbl,2] 
            # cesta nyni konci drive, nez v novy 
            i_novy = np.where(hranaJeDostupnaBlokace)[0][0] + 1 
            novy = cesta[i_novy] 
            kandidati[a] = novy
            cesta = cesta[:(i_novy+1)] 
            hranaJeDostupnaBlokace = hranaJeDostupnaBlokace[:i_novy] 
            dobaJizdy = pm.delkaTrasy(casyCG,cesta) 
        else: 
            dobaOprav = 0 
            dobaJizdy = pm.delkaTrasy(casyCG,cesta) 
        char[a] = 1 / (odpracovaneDoby[a] + dobaJizdy + dobaOprav) 
    if all(char==0): 
        return (None, None)  # nenastane, dokud pocetKomponent > 1 
    else: 
        char = char / sum(char) 
        cchar = np.cumsum(char) 
        b = np.where(cchar > np.random.rand())[0][0]  # vitez 
        return (b, kandidati[b]) 

def vyberemeMravence2(odpracovaneDoby): 
    """ Vyber mravence k presunu podle M. Trneny, 2022 
        Nejmene vytizeny mravenec. """ 
    return np.argmin(odpracovaneDoby) 



def konstrukceTras(startyCG:np.ndarray, poctyTymuVeStanicich:np.ndarray, 
                   feromony:np.ndarray, parametry:dict, casyCG:np.ndarray, 
                   blokaceCG:np.ndarray, hranyKMG:np.ndarray): 
    """
        Vstupy: 
            startyCG... index stanic v uplnem grafu CG 
            poctyTymuVeStanicich... vektor stejne delky jako startyCG (nezapor. int) 
            feromony... matice aktualniho mnozstvi feromonu na hranach 
            parametry... slovnik obsahujici m.j. feromony, alfa, beta 
                alfa... mocnitel feromonove casti vyberu souseda 
                beta... mocnitel heuristicke casti vyberu souseda 
            casyCG... matice dob jizd z uzluCG do uzluCG, rozmery stejne jako feromony 
            blokaceCG... (3 sloupce) zablokovane hrany v CG s dobou opravy (sekund) 
            hranyKMG... hrany spojujici komponenty, hr. ohod. dobou oprav 
        Prace: 
            Jedna m-tice mravenecu (ceta cestaru) sestroji jedno reseni OPRBRN. 
            Tak probehne jedna sezona. Zde je uzit pristup 6. Viz poznamky.txt 
            nebo primo kod teto funkce. 
        Vystupy: 
            reseni... np.ndarray, dvojice (s1,s2, s3) seznamu s1, s2, s3. 
            len(s1) = len(s2)+1. len(s3)<=len(s1). 
                s1... seznam m-tice tras. Trasa je seznam uzlu CG, 
                    Trasa zacina i konci ve stejnem uzlu. Mezi sousednimi uzly 
                    trasy neni uz zadny jiny uzel CG. 
                s2... seznam m-tice vektoru oprav odpovidajici trasy. 
                    Vektor je o 1 kratsi, nez prislusna trasa, znaci totiz jeji hrany. 
                    Znaci je hodnotami 1 = oprava, 0 = pouhy prujezd. 
                s3... seznam m-tice seznamu CG uzlu, ktere dany mravenec volil. 
                    Doplnenim CG uzlu nejkratsi cesty v CG mezi dvema sousednimi 
                    uzly seznamu s3[i], (i in (0:(m-1))), vznikne trasa s1[i]. 
    """
    
    alfa = parametry["alfa"] 
    beta = parametry["beta"] 
    zpvm = parametry["zpusobVyberuMravence"] 
    
    # inicializace: 
    m = sum(poctyTymuVeStanicich) 
    s = len(poctyTymuVeStanicich)  # pocet stanic = len(startyCG) 
    Ncg = casyCG.shape[0]    # pocet uzlu cg 
    dostupneUzly = np.arange(Ncg).tolist() 
    odpracovaneDoby = np.array([0]*m) 
    trasy = []          # seznam seznamuu navstivenych uzlu CG 
    seznamSeznamuOprav = [] 
    mravenciVolby = []  # seznamy seznamuu uzlu CG, ktere voli mravenec 
    seznamyDostupnychUzlu = []  # ke kazde stanici jeden seznam 
    trasyKodyHranCG = [] 
    zacatkyVsechTras = [] 
    indexStaniceMravence = [] 
    for i in range(s):   # starty (stanice) 
        tmpl = [startyCG[i]] * poctyTymuVeStanicich[i] 
        zacatkyVsechTras.extend(tmpl) 
        tmpl = [i] * poctyTymuVeStanicich[i] 
        indexStaniceMravence.extend(tmpl) 
        dou = dostupneUzly.copy() 
        dou.remove(startyCG[i]) 
        seznamyDostupnychUzlu.append(dou) 
    for i in range(m):   # mravenci (tymy) 
        trasy.append([int(zacatkyVsechTras[i])]) 
        seznamSeznamuOprav.append([]) 
        mravenciVolby.append([int(zacatkyVsechTras[i])]) 
        trasyKodyHranCG.append([]) 
    for i in range(s): 
        if not(i in indexStaniceMravence): 
            seznamyDostupnychUzlu[i] = [] 
    del dostupneUzly, dou 
    
    # mravenci budou postupne ujidat... 
    seznamDostupnychBlokaci = blokaceCGkody.copy() 
    paryZnacek = hranyKMG[:,0:2].copy() 
    uzlyCGkomponenty = legendaCG[:,2].copy() 
    komponenty = np.unique(uzlyCGkomponenty) 
    pocetKomponent = len(komponenty) 
    
    # Prilezitostne napojovani komponent 
    while (pocetKomponent > 1):  # k. rozpojeny 
        
        # 1) Vybrany mravenec vybere uzel z jemu dostupnych uzlu. 
        if (zpvm == 'vm1'): 
            dv = vyberemeMravence1(trasy, odpracovaneDoby, feromony, alfa, beta, 
                             indexStaniceMravence, seznamyDostupnychUzlu, 
                             seznamDostupnychBlokaci) 
            ukaz = dv[0]   # ne None ? 
            aktualni = trasy[ukaz][-1] 
            novy = dv[1]   # ne None ? 
        else:  # (zpvm == 'vm2'): 
            ukaz = vyberemeMravence2(odpracovaneDoby) 
            aktualni = trasy[ukaz][-1] 
            iukaz = indexStaniceMravence[ukaz] 
            sousede = seznamyDostupnychUzlu[iukaz] 
            if len(sousede)>0: 
                f = feromony[aktualni,sousede] 
                f = np.power(f, alfa) 
                h = casyCG[aktualni,sousede] 
                h = 1 / h 
                h = np.power(h, beta) 
                psti = f*h / f.dot(h) 
                cpsti = np.cumsum(psti) 
                # (ruletove kolo) vybere hranu a nasledujici uzel. Slo by lepe. 
                vyber = np.where(cpsti > np.random.rand())[0][0] 
                novy = sousede[vyber] 
            else:  # mozny jedine navrat 
                novy = trasy[ukaz][0] 
        
        # 2) Na tehle hrane (CG ceste) opravi prvni dostupnou blokaci. 
        # dostupne = neopravene a spojujici dve ruzne komponenty 
        cesta = cestyCG[aktualni][novy] 
        hranyCesty = np.vstack((np.array(cesta[0:-1]), np.array(cesta[1:]))) 
        hranyCesty = hranyCesty.transpose() 
        kodyCGhranCesty = pm.kodyzHranCG(hranyCesty) 
        #hranaJeBlokace = np.in1d(kodyCGhranCesty, blokaceCGkody) 
        hranaJeDostupnaBlokace = np.in1d(
            kodyCGhranCesty, seznamDostupnychBlokaci) 
        
        # 3) Zmeni se odpracovana doba mravence a znacky komponent. 
        if any(hranaJeDostupnaBlokace): 
            opravovaneBlokace = kodyCGhranCesty[hranaJeDostupnaBlokace] 
            opravovanaBlokace = opravovaneBlokace[0]  # jen prvni 
            ktbl = np.in1d(blokaceCGkody, opravovanaBlokace) 
            casOpravy = blokaceCG[ktbl,2] 
            # cesta nyni konci drive, nez v novy 
            i_novy = np.where(hranaJeDostupnaBlokace)[0][0] + 1 
            novy = cesta[i_novy] 
            cesta = cesta[:(i_novy+1)] 
            hranaJeDostupnaBlokace = hranaJeDostupnaBlokace[:i_novy] 
        else: 
            casOpravy = 0 
        casJizdy = sum(casyCG[cesta[0:-1],cesta[1:]]) 
        odpracovaneDoby[ukaz] += (casJizdy + casOpravy) 
        trasy[ukaz].extend(cesta[1:]) 
        seznamSeznamuOprav[ukaz].extend((1*hranaJeDostupnaBlokace).tolist()) 
        mravenciVolby[ukaz].append(novy) 
        
        # Zjistim znacku prvniho uzlu cesty a touhle znackou 
        # oznacim uzly vsech komponent, kterymi cesta prochazi. 
        znacka1 = uzlyCGkomponenty[cesta[0]]
        for i in range(len(cesta)): 
            znacka2 = uzlyCGkomponenty[cesta[i]] 
            ostrovane = legendaCG[uzlyCGkomponenty == znacka2,1] 
            uzlyCGkomponenty[ostrovane] = znacka1  # preznaceni napojenych 
        
        # Nove zvoleny uzel at tento mravenec priste nevoli. 
        # seznamyDostupnychUzlu[index_stanice].remove(novy) 
        # A at tento mravenec nevoli ani uzly v stejne komponente. 
        index_stanice = indexStaniceMravence[ukaz] 
        ostrovane = legendaCG[uzlyCGkomponenty == uzlyCGkomponenty[novy],1] 
        for o in ostrovane: 
            if (o in seznamyDostupnychUzlu[index_stanice]): 
                seznamyDostupnychUzlu[index_stanice].remove(o) 
        
        # aktualizujeme pary znacek (sousedicich komponent) 
        for kom in komponenty: 
            novaKom = uzlyCGkomponenty[legendaCG[:,2]==kom][0] 
            paryZnacek[paryZnacek==kom] = novaKom 
        # ponechame blokace, ktere maji ruzne znacene konce 
        seznamDostupnychBlokaci = \
            blokaceCGkody[paryZnacek[:,0]!=paryZnacek[:,1]].copy() 
        
        komponenty = np.unique(uzlyCGkomponenty) 
        pocetKomponent = len(komponenty) 
    
    # uzavreni tras (navrat do stanic): 
    for i in range(m): 
        if (trasy[i][-1] != trasy[i][0]): 
            cesta = cestyCG[trasy[i][-1]][trasy[i][0]] 
            trasy[i].extend(cesta[1:]) 
            opravy = [0]*(len(cesta)-1)
            seznamSeznamuOprav[i].extend(opravy) 
            mravenciVolby[i].append(trasy[i][0]) 
    reseni = (trasy, seznamSeznamuOprav, mravenciVolby) 
    return reseni 



def aktualizaceMeziFeromonu(reseni, odpar, pomer): 
    feroMax = 1 / (odpar * sum(cenyTras(reseni))) 
    feroMin = feroMax / pomer  # Papir 2023-02-14 
    return (feroMin, feroMax) 

def aktualizaceFeromonu(feromony:np.ndarray, feroMin, feroMax, 
                        parametry:dict, reseni:tuple): 
    """
    Vstupy: 
        feromony... ctvercova matice feromonu 
        feroMin... hodnota, pod kterou se mnozstvi feromonu na hrane nedostane 
        feroMax... hodnota, kterou mnozstvi feromonu na hrane neprekroci 
        parametry... slovnik obsahujici m.j. odpar, feroMax, feroMin 
            odpar... cislo v otevrenem intervalu (0,1). 
        reseni... trojice (typ tuple), ktera je vysledkem konstrukce tras 
    Prace: 
        Na hranu dame feromon jen jednou, i kdyby ji 
        v jednom reseni pouzilo vice mravencu. 
    Vystupy: 
        feromony... ctvercova matice s novymi hodnotami feromonu. 
    """ 
    feromony = feromony.copy() 
    odpar = parametry["odpar"] 
    # Max-Min system - odpar: 
    feromony[feromony > feroMax] = feroMax 
    feromony = feromony * (1 - odpar) 
    feromony[feromony < feroMin] = feroMin 
    hrany = set() 
    trasy = reseni[2]   # mravenci volby 
    for trasa in trasy: 
        for i in range(len(trasa)-1): 
            hrany |= {pm.kodzHranyCG([trasa[i],trasa[i+1]])} 
    hrany = np.array(list(hrany)) 
    hrany = pm.hranyCGzkodu(hrany) 
    # Max-Min system - zanechani stop: 
    delta = 1 / sum(cenyTras(reseni)) 
    for i in range(hrany.shape[0]): 
        feromony[hrany[i,0],hrany[i,1]] += delta 
        feromony[hrany[i,1],hrany[i,0]] += delta 
    feromony[feromony < feroMin] = feroMin 
    feromony[feromony > feroMax] = feroMax 
    # korekce na diagonale 
    xx = np.arange(feromony.shape[0]) 
    feromony[xx,xx] = feroMin 
    return feromony 



### Hlavni funkce 
def MMAS(casyCG:np.ndarray, startyCG:np.ndarray, poctyTymuVeStanicich:np.ndarray, 
          parametry:dict, zaznamnik:list): 
    
    Ncg = casyCG.shape[0] 
    praktickyNekonecno = np.max(casyCG)*Ncg*Ncg + np.sum(blokaceCG[:,2]) 
    iterLimit = parametry["iterLimit"] 
    iterLimitKopt = parametry["iterLimitKopt"] 
    kopt = parametry["zlepsovaciAlgoritmus"]
    frekvencePruzkumu = parametry["frekvencePruzkumu"] 
    uzitPocatecniReseni = parametry["uzitPocatecniReseni"]
    m = parametry["mravencu"] 
    pomerFmaxFmin = parametry["pomerFmaxFmin"]
    feroMin = parametry["feroMin"] 
    feroMax = parametry["feroMax"] 
    feromony = parametry["feromony"] 
    odpar = parametry["odpar"] 
    typCeny =  parametry["typCeny"] 
    reseni_bs = None 
    reseni_bs_cena = np.array([praktickyNekonecno]*m) 
    pocetKolonii = parametry["pocetKolonii"] 
    aktualizovatFeromony = parametry["aktualizovatFeromony"] 
    
    if uzitPocatecniReseni: 
        # Vypocet pocatecniho reseni: 
        kostry = pm.levneKostry(multigraf=Mgr, pocetKoster=1) 
        kostraKMG = kostry[0] 
        navstivit = blokaceCG[kostraKMG[:,3],0:2].tolist() 
        pocatecniReseni = \
            reseniA.algoUspor3(casyCG, startyCG, poctyTymuVeStanicich, navstivit) 
        pocatecniReseni = adapterReseniAZ_M(pocatecniReseni) 
        # pred zacatkem iteraci mravenci pujdou po trasach z reseniA 
        # a zanechaji feromony na hranach pocatecniho reseni. 
        (feroMin, feroMax) = aktualizaceMeziFeromonu(
            pocatecniReseni, odpar, pomerFmaxFmin) 
        feromony = aktualizaceFeromonu(
            feromony, feroMin, feroMax, parametry, pocatecniReseni) 
        reseni_bs = pocatecniReseni      # Best So far solution 
        reseni_bs_cena = cenyTras(pocatecniReseni) 
        reseni_ib = pocatecniReseni      # Iteration Best solution 
        reseni_ib_cena = cenyTras(pocatecniReseni) 
    
    # Ted iterace: 
    for i in range(iterLimit): 
        reseni_ib = None 
        reseni_ib_cena = np.array([praktickyNekonecno]*m) 
        for j in range(pocetKolonii): 
            reseni = konstrukceTras(startyCG, poctyTymuVeStanicich, feromony, 
                                parametry, casyCG, blokaceCG, hranyKMG) 
            if (kopt != 'nic'): 
                reseni = reseniZ.zlepsi(reseni, algoritmus=kopt, limit=iterLimitKopt) 
                reseni = adapterReseniAZ_M(reseni) 
            reseni_cena = cenyTras(reseni) 
            if (typCeny == 'vse'): 
                cena = pm.mensiLexikograficky(reseni_cena, reseni_bs_cena) 
                if (all(cena == reseni_cena)):    #  reseni_cena  L<L  reseni_bs_cena 
                    reseni_bs = reseni 
                    reseni_bs_cena = reseni_cena 
                cena = pm.mensiLexikograficky(reseni_cena, reseni_ib_cena) 
                if (all(cena == reseni_cena)):
                    reseni_ib = reseni 
                    reseni_ib_cena = reseni_cena 
            elif (typCeny == 'sum'): 
                rc = pm.cenaQ1(reseni_cena) 
                rbc = pm.cenaQ1(reseni_bs_cena) 
                ric = pm.cenaQ1(reseni_ib_cena) 
                if (rc < rbc): 
                    reseni_bs = reseni 
                    reseni_bs_cena = reseni_cena 
                if (rc < ric): 
                    reseni_ib = reseni 
                    reseni_ib_cena = reseni_cena 
            else: # (typCeny == 'max'): 
                rc = pm.cenaQ2(reseni_cena) 
                rbc = pm.cenaQ2(reseni_bs_cena) 
                ric = pm.cenaQ2(reseni_ib_cena) 
                if (rc < rbc): 
                    reseni_bs = reseni 
                    reseni_bs_cena = reseni_cena 
                if (rc < ric): 
                    reseni_ib = reseni 
                    reseni_ib_cena = reseni_cena 
        # zaznamnik urcen pro zapis vystupu kazde iterace 
        reseni_ib_hrube = reseniZ.adapterReseniMZ(reseni_ib) 
        zaznamnik.append(reseni_ib_hrube) 
        if aktualizovatFeromony: 
            ibbs = frekvencePruzkumu(i) # (~ cetnost ib) 
            if np.random.random() <= ibbs: 
                trasy = reseni_ib 
            else: 
                trasy = reseni_bs 
            (feroMin, feroMax) = aktualizaceMeziFeromonu(
                reseni_bs, odpar, pomerFmaxFmin) 
            # zmenime stavy prirustku feromonu na hranach 
            feromony = aktualizaceFeromonu(
                feromony, feroMin, feroMax, parametry, trasy) 
        reseni_bs_hrube = reseniZ.adapterReseniMZ(reseni_bs) 
    return reseni_bs_hrube 
# MMAS() konec 

