


import numpy as np 
import pandas as pd 
import networkx as nx 

# VSTUPNI UDAJE (lze menit podle potreby): 
sit = "C:\\Prilohy\\Ulohy\\sit003"
situace = "k06d02_test"
# Adresare musi byt ve tvaru:   sit\situace, napriklad:  
# C:\Prilohy\Ulohy\sit003\k06d02_test 

def nahrajZadani(sit, situace): 
    # Pozice uzluu v rovine 
    xy = pd.read_csv(sit+"\\xy.txt", header="infer", sep=" ")
    xy = xy.to_numpy() 
    # Uzly puvodni site s poctem obyvatel 
    uzly = pd.read_csv(sit+"\\uzly.txt", header="infer", sep=" ")
    uzly = uzly.to_numpy() 
    # Hrany puvodni site s trvanim presunu po hrane 
    hrany = pd.read_csv(sit+"\\hrany.txt", header="infer", sep=" ")
    hrany = hrany.to_numpy() 
    # Ktere hrany jsou zablokovane 
    blokace = pd.read_csv(sit+"\\"+situace+"\\blokace.txt", header=None)
    blokace = blokace.to_numpy()
    # stanice 
    starty = pd.read_csv(sit+"\\"+situace+"\\starty.txt", header=None)
    starty = starty.to_numpy()
    return (xy, uzly, hrany, blokace, starty) 

# Puvodni sit a rozpadla sit: 
(xy, uzly, hrany, blokace, starty) = nahrajZadani(sit, situace)
puvodniSit = nx.Graph() 
rozpadlaSit = nx.Graph() 
uattr = []
hattrse = []
hattrse2 = []
for i in range(xy.shape[0]): 
    vlastnosti = {"xy": xy[i,:], "obyvatel": uzly[i,1]} 
    uattr.append((i, vlastnosti))
for i in range(0,len(hrany)): 
    prujezdna = not(i in blokace) 
    vlastnosti = {"sekund": hrany[i,2], "prujezdna": prujezdna, 
                  "id_hrany": i} 
    hattrse.append((hrany[i,0], hrany[i,1], vlastnosti)) 
    if prujezdna: 
        vlastnosti = {"sekund": hrany[i,2], "id_hrany": i} 
        hattrse2.append((hrany[i,0], hrany[i,1], vlastnosti)) 
puvodniSit.add_nodes_from(uattr)
puvodniSit.add_edges_from(hattrse) 
rozpadlaSit.add_nodes_from(uattr)
rozpadlaSit.add_edges_from(hattrse2) 
del uattr, hattrse, hattrse2 



def generujPrislusnostUzluKomponentam(): 
    S = [rozpadlaSit.subgraph(c).copy() for c in nx.connected_components(rozpadlaSit)] 
    uzlyKomponenty = [] 
    for i in range(len(S)): 
        for node in S[i].nodes: 
            uzlyKomponenty.append([int(node), i])  # [id uzlu, cislo komponenty] 
    uzlyKomponenty = np.array(uzlyKomponenty) 
    return uzlyKomponenty 

def generujCasyCG(uzlyCG): 
    Ncg = len(uzlyCG.ravel()) 
    casyCG = np.zeros((Ncg,Ncg), dtype=int) 
    for i in range(Ncg): 
        od = uzlyCG[i,0] 
        dd = nx.shortest_path_length(puvodniSit, source=od, target=None, 
                                weight="sekund", method="dijkstra") 
        for do in (uzlyCG.ravel().tolist()): 
            delka = dd[do] 
            j = np.where(uzlyCG.ravel() == do)[0][0] 
            casyCG[i,j] = delka 
            casyCG[j,i] = delka 
    return casyCG 



# toto je hlavni funkce 
def vytvorCGdata(uzly, hrany, blokace, starty): 
    # Od hran odstranim ty zablokovane a z pozustalych sestrojim nesouvisly networkx. 
    uzlyKomponenty = generujPrislusnostUzluKomponentam() 
    # V nesouvislem networkx najdu vsechny komponenty, uzlum v komponente priradim stejne cislo, vznikne uKompId. 
    uKompId = uzlyKomponenty[uzlyKomponenty[:,0].ravel().argsort(), 1].copy() 
    uKompId = uKompId.reshape(len(uKompId),1) 
    # uzlyCG jsou ty incidentni se zablokovanymi hranami sjednoceno se starty. 
    # uzlyCG jsou nektere indexy do matice uzly. 
    uzlyCG = set(np.unique(hrany[blokace.ravel(),:2])) | set(starty.ravel().tolist()) 
    uzlyCG = np.array(list(uzlyCG)).reshape(len(uzlyCG),1) 
    # V puvodnim souvislem networkx najdeme delky nejkratsich cest mezi uzlyCG a udelame z toho matici casyCG. 
    casyCG = generujCasyCG(uzlyCG) 
    return (uKompId, uzlyCG, casyCG) 
# 

def ulozCGdata(uKompId, uzlyCG, casyCG): 
    outcesta = sit+"\\"+situace+"\\"+ "uKompId.txt" 
    outtabulka = pd.DataFrame(uKompId) 
    outtabulka.to_csv(outcesta, header=False, index=False) 
    outcesta = sit+"\\"+situace+"\\"+ "uzlyCG.txt" 
    outtabulka = pd.DataFrame(uzlyCG, columns=['id']) 
    outtabulka.to_csv(outcesta, header=True, index=False) 
    outcesta = sit+"\\"+situace+"\\"+ "casyCG.txt" 
    outtabulka = pd.DataFrame(casyCG) 
    outtabulka.to_csv(outcesta, header=False, index=False) 
# 

# Tohle ulozime na spravne misto: 
(uKompId, uzlyCG, casyCG) = vytvorCGdata(uzly, hrany, blokace, starty) 
ulozCGdata(uKompId, uzlyCG, casyCG) 
# Hotovo. 





